@@ -19,31 +19,36 @@ import Errors
 import Parsing
 import Version
 from Scanning import PyrexScanner, FileSourceDescriptor
-from Errors import PyrexError, CompileError, error
+from Errors import PyrexError, CompileError, InternalError, error
 from Symtab import BuiltinScope, ModuleScope
 from Cython import Utils
+from Cython.Utils import open_new_file, replace_suffix
+import CythonScope
 
 module_name_pattern = re.compile(r"[A-Za-z_][A-Za-z0-9_]*(\.[A-Za-z_][A-Za-z0-9_]*)*$")
 
-# Note: PHASES and TransformSet should be removed soon; but that's for
-# another day and another commit.
-PHASES = [
-    'before_analyse_function', # run in FuncDefNode.generate_function_definitions
-    'after_analyse_function'   # run in FuncDefNode.generate_function_definitions
-]
-
-class TransformSet(dict):
-    def __init__(self):
-        for name in PHASES:
-            self[name] = []
-    def run(self, name, node, **options):
-        assert name in self, "Transform phase %s not defined" % name
-        for transform in self[name]:
-            transform(node, phase=name, **options)
-
 verbose = 0
 
-class Context:
+def dumptree(t):
+    # For quick debugging in pipelines
+    print t.dump()
+    return t
+
+class CompilationData(object):
+    #  Bundles the information that is passed from transform to transform.
+    #  (For now, this is only)
+
+    #  While Context contains every pxd ever loaded, path information etc.,
+    #  this only contains the data related to a single compilation pass
+    #
+    #  pyx                   ModuleNode              Main code tree of this compilation.
+    #  pxds                  {string : ModuleNode}   Trees for the pxds used in the pyx.
+    #  codewriter            CCodeWriter             Where to output final code.
+    #  options               CompilationOptions
+    #  result                CompilationResult
+    pass
+
+class Context(object):
     #  This class encapsulates the context needed for compiling
     #  one or more Cython implementation files along with their
     #  associated and imported declaration files. It includes
@@ -54,13 +59,151 @@ class Context:
     #  include_directories   [string]
     #  future_directives     [object]
     
-    def __init__(self, include_directories):
+    def __init__(self, include_directories, pragma_overrides):
         #self.modules = {"__builtin__" : BuiltinScope()}
-        import Builtin
+        import Builtin, CythonScope
         self.modules = {"__builtin__" : Builtin.builtin_scope}
+        self.modules["cython"] = CythonScope.create_cython_scope(self)
         self.include_directories = include_directories
         self.future_directives = set()
-        
+        self.pragma_overrides = pragma_overrides
+
+        self.pxds = {} # full name -> node tree
+
+        standard_include_path = os.path.abspath(
+            os.path.join(os.path.dirname(__file__), '..', 'Includes'))
+        self.include_directories = include_directories + [standard_include_path]
+
+    def create_pipeline(self, pxd, py=False):
+        from Visitor import PrintTree
+        from ParseTreeTransforms import WithTransform, NormalizeTree, PostParse, PxdPostParse
+        from ParseTreeTransforms import AnalyseDeclarationsTransform, AnalyseExpressionsTransform
+        from ParseTreeTransforms import CreateClosureClasses, MarkClosureVisitor, DecoratorTransform
+        from ParseTreeTransforms import InterpretCompilerDirectives, TransformBuiltinMethods
+        from ParseTreeTransforms import AlignFunctionDefinitions
+        from AutoDocTransforms import EmbedSignature
+        from Optimize import FlattenInListTransform, SwitchTransform, IterationTransform
+        from Optimize import FlattenBuiltinTypeCreation, ConstantFolding, FinalOptimizePhase
+        from Buffer import IntroduceBufferAuxiliaryVars
+        from ModuleNode import check_c_declarations
+
+        # Temporary hack that can be used to ensure that all result_code's
+        # are generated at code generation time.
+        import Visitor
+        class ClearResultCodes(Visitor.CythonTransform):
+            def visit_ExprNode(self, node):
+                self.visitchildren(node)
+                node.result_code = "<cleared>"
+                return node
+
+        if pxd:
+            _check_c_declarations = None
+            _specific_post_parse = PxdPostParse(self)
+        else:
+            _check_c_declarations = check_c_declarations
+            _specific_post_parse = None
+            
+        if py and not pxd:
+            _align_function_definitions = AlignFunctionDefinitions(self)
+        else:
+            _align_function_definitions = None
+ 
+        return [
+            NormalizeTree(self),
+            PostParse(self),
+            _specific_post_parse,
+            InterpretCompilerDirectives(self, self.pragma_overrides),
+            _align_function_definitions,
+            FlattenInListTransform(),
+            WithTransform(self),
+            DecoratorTransform(self),
+            AnalyseDeclarationsTransform(self),
+            EmbedSignature(self),
+            TransformBuiltinMethods(self),
+            IntroduceBufferAuxiliaryVars(self),
+            _check_c_declarations,
+            AnalyseExpressionsTransform(self),
+            FlattenBuiltinTypeCreation(),
+            ConstantFolding(),
+            IterationTransform(),
+            SwitchTransform(),
+            FinalOptimizePhase(self),
+#            ClearResultCodes(self),
+#            SpecialFunctions(self),
+            #        CreateClosureClasses(context),
+            ]
+
+    def create_pyx_pipeline(self, options, result, py=False):
+        def generate_pyx_code(module_node):
+            module_node.process_implementation(options, result)
+            result.compilation_source = module_node.compilation_source
+            return result
+
+        def inject_pxd_code(module_node):
+            from textwrap import dedent
+            stats = module_node.body.stats
+            for name, (statlistnode, scope) in self.pxds.iteritems():
+                # Copy over function nodes to the module
+                # (this seems strange -- I believe the right concept is to split
+                # ModuleNode into a ModuleNode and a CodeGenerator, and tell that
+                # CodeGenerator to generate code both from the pyx and pxd ModuleNodes.
+                 stats.append(statlistnode)
+                 # Until utility code is moved to code generation phase everywhere,
+                 # we need to copy it over to the main scope
+                 module_node.scope.utility_code_list.extend(scope.utility_code_list)
+            return module_node
+
+        return ([
+                create_parse(self),
+            ] + self.create_pipeline(pxd=False, py=py) + [
+                inject_pxd_code,
+                generate_pyx_code,
+            ])
+
+    def create_pxd_pipeline(self, scope, module_name):
+        def parse_pxd(source_desc):
+            tree = self.parse(source_desc, scope, pxd=True,
+                              full_module_name=module_name)
+            tree.scope = scope
+            tree.is_pxd = True
+            return tree
+
+        from CodeGeneration import ExtractPxdCode
+
+        # The pxd pipeline ends up with a CCodeWriter containing the
+        # code of the pxd, as well as a pxd scope.
+        return [parse_pxd] + self.create_pipeline(pxd=True) + [
+            ExtractPxdCode(self),
+            ]
+            
+    def create_py_pipeline(self, options, result):
+        return self.create_pyx_pipeline(options, result, py=True)
+
+
+    def process_pxd(self, source_desc, scope, module_name):
+        pipeline = self.create_pxd_pipeline(scope, module_name)
+        result = self.run_pipeline(pipeline, source_desc)
+        return result
+    
+    def nonfatal_error(self, exc):
+        return Errors.report_error(exc)
+
+    def run_pipeline(self, pipeline, source):
+        err = None
+        data = source
+        try:
+            for phase in pipeline:
+                if phase is not None:
+                    data = phase(data)
+        except CompileError, err:
+            # err is set
+            Errors.report_error(err)
+        except InternalError, err:
+            # Only raise if there was not an earlier error
+            if Errors.num_errors == 0:
+                raise
+        return (err, data)
+
     def find_module(self, module_name, 
             relative_to = None, pos = None, need_pxd = 1):
         # Finds and returns the module scope corresponding to
@@ -74,10 +217,13 @@ class Context:
         if debug_find_module:
             print("Context.find_module: module_name = %s, relative_to = %s, pos = %s, need_pxd = %s" % (
                     module_name, relative_to, pos, need_pxd))
+
         scope = None
         pxd_pathname = None
         if not module_name_pattern.match(module_name):
-            raise CompileError((path, 0, 0),
+            if pos is None:
+                pos = (module_name, 0, 0)
+            raise CompileError(pos,
                 "'%s' is not a valid module name" % module_name)
         if "." not in module_name and relative_to:
             if debug_find_module:
@@ -107,15 +253,21 @@ class Context:
                 if debug_find_module:
                     print("......found ", pxd_pathname)
                 if not pxd_pathname and need_pxd:
-                    error(pos, "'%s.pxd' not found" % module_name)
+                    package_pathname = self.search_include_directories(module_name, ".py", pos)
+                    if package_pathname and package_pathname.endswith('__init__.py'):
+                        pass
+                    else:
+                        error(pos, "'%s.pxd' not found" % module_name)
             if pxd_pathname:
                 try:
                     if debug_find_module:
                         print("Context.find_module: Parsing %s" % pxd_pathname)
                     source_desc = FileSourceDescriptor(pxd_pathname)
-                    pxd_tree = self.parse(source_desc, scope, pxd = 1,
-                                          full_module_name = module_name)
-                    pxd_tree.analyse_declarations(scope)
+                    err, result = self.process_pxd(source_desc, scope, module_name)
+                    if err:
+                        raise err
+                    (pxd_codenodes, pxd_scope) = result
+                    self.pxds[module_name] = (pxd_codenodes, pxd_scope)
                 except CompileError:
                     pass
         return scope
@@ -123,6 +275,11 @@ class Context:
     def find_pxd_file(self, qualified_name, pos):
         # Search include path for the .pxd file corresponding to the
         # given fully-qualified module name.
+        # Will find either a dotted filename or a file in a
+        # package directory. If a source file position is given,
+        # the directory containing the source file is searched first
+        # for a dotted filename, and its containing package root
+        # directory is searched first for a non-dotted filename.
         return self.search_include_directories(qualified_name, ".pxd", pos)
 
     def find_pyx_file(self, qualified_name, pos):
@@ -134,27 +291,30 @@ class Context:
         # Search list of include directories for filename.
         # Reports an error and returns None if not found.
         path = self.search_include_directories(filename, "", pos,
-                                               split_package=False)
+                                               include=True)
         if not path:
             error(pos, "'%s' not found" % filename)
         return path
     
     def search_include_directories(self, qualified_name, suffix, pos,
-                                   split_package=True):
+                                   include=False):
         # Search the list of include directories for the given
         # file name. If a source file position is given, first
         # searches the directory containing that file. Returns
         # None if not found, but does not report an error.
+        # The 'include' option will disable package dereferencing.
         dirs = self.include_directories
         if pos:
             file_desc = pos[0]
             if not isinstance(file_desc, FileSourceDescriptor):
                 raise RuntimeError("Only file sources for code supported")
-            here_dir = os.path.dirname(file_desc.filename)
-            dirs = [here_dir] + dirs
+            if include:
+                dirs = [os.path.dirname(file_desc.filename)] + dirs
+            else:
+                dirs = [self.find_root_package_dir(file_desc.filename)] + dirs
 
         dotted_filename = qualified_name + suffix
-        if split_package:
+        if not include:
             names = qualified_name.split('.')
             package_names = names[:-1]
             module_name = names[-1]
@@ -165,7 +325,7 @@ class Context:
             path = os.path.join(dir, dotted_filename)
             if os.path.exists(path):
                 return path
-            if split_package:
+            if not include:
                 package_dir = self.check_package_dir(dir, package_names)
                 if package_dir is not None:
                     path = os.path.join(package_dir, module_filename)
@@ -177,15 +337,27 @@ class Context:
                         return path
         return None
 
+    def find_root_package_dir(self, file_path):
+        dir = os.path.dirname(file_path)
+        while self.is_package_dir(dir):
+            parent = os.path.dirname(dir)
+            if parent == dir:
+                break
+            dir = parent
+        return dir
+
+    def is_package_dir(self, dir):
+        package_init = os.path.join(dir, "__init__.py")
+        return os.path.exists(package_init) or \
+            os.path.exists(package_init + "x") # same with .pyx
+
     def check_package_dir(self, dir, package_names):
         package_dir = os.path.join(dir, *package_names)
         if not os.path.exists(package_dir):
             return None
         for dirname in package_names:
             dir = os.path.join(dir, dirname)
-            package_init = os.path.join(dir, "__init__.py")
-            if not os.path.exists(package_init) and \
-                    not os.path.exists(package_init + "x"): # same with .pyx ?
+            if not self.is_package_dir(dir):
                 return None
         return package_dir
 
@@ -261,6 +433,8 @@ class Context:
             finally:
                 f.close()
         except UnicodeDecodeError, msg:
+            #import traceback
+            #traceback.print_exc()
             error((source_desc, 0, 0), "Decoding error, missing or incorrect coding=<encoding-name> at top of source (%s)" % msg)
         if Errors.num_errors > 0:
             raise CompileError
@@ -285,55 +459,25 @@ class Context:
         names.reverse()
         return ".".join(names)
 
-    def compile(self, source, options = None, full_module_name = None):
-        # Compile a Pyrex implementation file in this context
-        # and return a CompilationResult.
-        if not options:
-            options = default_options
-        result = CompilationResult()
-        cwd = os.getcwd()
-
-        source = os.path.join(cwd, source)
-        result.main_source_file = source
-
+    def setup_errors(self, options):
         if options.use_listing_file:
             result.listing_file = Utils.replace_suffix(source, ".lis")
             Errors.open_listing_file(result.listing_file,
                 echo_to_stderr = options.errors_to_stderr)
         else:
             Errors.open_listing_file(None)
-        if options.output_file:
-            result.c_file = os.path.join(cwd, options.output_file)
-        else:
-            if options.cplus:
-                c_suffix = ".cpp"
-            else:
-                c_suffix = ".c"
-            result.c_file = Utils.replace_suffix(source, c_suffix)
-        c_stat = None
-        if result.c_file:
-            try:
-                c_stat = os.stat(result.c_file)
-            except EnvironmentError:
-                pass
-        full_module_name = full_module_name or self.extract_module_name(source, options)
-        source = FileSourceDescriptor(source)
-        initial_pos = (source, 1, 0)
-        scope = self.find_module(full_module_name, pos = initial_pos, need_pxd = 0)
-        errors_occurred = False
-        try:
-            tree = self.parse(source, scope, pxd = 0,
-                              full_module_name = full_module_name)
-            tree.process_implementation(scope, options, result)
-        except CompileError:
-            errors_occurred = True
+
+    def teardown_errors(self, err, options, result):
+        source_desc = result.compilation_source.source_desc
+        if not isinstance(source_desc, FileSourceDescriptor):
+            raise RuntimeError("Only file sources for code supported")
         Errors.close_listing_file()
         result.num_errors = Errors.num_errors
         if result.num_errors > 0:
-            errors_occurred = True
-        if errors_occurred and result.c_file:
+            err = True
+        if err and result.c_file:
             try:
-                Utils.castrate_file(result.c_file, os.stat(source.filename))
+                Utils.castrate_file(result.c_file, os.stat(source_desc.filename))
             except EnvironmentError:
                 pass
             result.c_file = None
@@ -346,7 +490,58 @@ class Context:
                     extra_objects = options.objects,
                     verbose_flag = options.show_version,
                     cplus = options.cplus)
-        return result
+
+def create_parse(context):
+    def parse(compsrc):
+        source_desc = compsrc.source_desc
+        full_module_name = compsrc.full_module_name
+        initial_pos = (source_desc, 1, 0)
+        scope = context.find_module(full_module_name, pos = initial_pos, need_pxd = 0)
+        tree = context.parse(source_desc, scope, pxd = 0, full_module_name = full_module_name)
+        tree.compilation_source = compsrc
+        tree.scope = scope
+        tree.is_pxd = False
+        return tree
+    return parse
+
+def create_default_resultobj(compilation_source, options):
+    result = CompilationResult()
+    result.main_source_file = compilation_source.source_desc.filename
+    result.compilation_source = compilation_source
+    source_desc = compilation_source.source_desc
+    if options.output_file:
+        result.c_file = os.path.join(compilation_source.cwd, options.output_file)
+    else:
+        if options.cplus:
+            c_suffix = ".cpp"
+        else:
+            c_suffix = ".c"
+        result.c_file = Utils.replace_suffix(source_desc.filename, c_suffix)
+    return result
+
+def run_pipeline(source, options, full_module_name = None):
+    # Set up context
+    context = Context(options.include_path, options.pragma_overrides)
+
+    # Set up source object
+    cwd = os.getcwd()
+    source_desc = FileSourceDescriptor(os.path.join(cwd, source))
+    full_module_name = full_module_name or context.extract_module_name(source, options)
+    source = CompilationSource(source_desc, full_module_name, cwd)
+
+    # Set up result object
+    result = create_default_resultobj(source, options)
+    
+    # Get pipeline
+    if source_desc.filename.endswith(".py"):
+        pipeline = context.create_py_pipeline(options, result)
+    else:
+        pipeline = context.create_pyx_pipeline(options, result)
+
+    context.setup_errors(options)
+    err, enddata = context.run_pipeline(pipeline, source)
+    context.teardown_errors(err, options, result)
+    return result
 
 #------------------------------------------------------------------------
 #
@@ -354,7 +549,17 @@ class Context:
 #
 #------------------------------------------------------------------------
 
-class CompilationOptions:
+class CompilationSource(object):
+    """
+    Contains the data necesarry to start up a compilation pipeline for
+    a single compilation unit.
+    """
+    def __init__(self, source_desc, full_module_name, cwd):
+        self.source_desc = source_desc
+        self.full_module_name = full_module_name
+        self.cwd = cwd
+
+class CompilationOptions(object):
     """
     Options to the Cython compiler:
     
@@ -369,7 +574,7 @@ class CompilationOptions:
                                 defaults to true when recursive is true.
     verbose           boolean   Always print source names being compiled
     quiet             boolean   Don't print source names in recursive mode
-    transforms        Transform.TransformSet Transforms to use on the parse tree
+    pragma_overrides  dict      Overrides for pragma options (see Options.py)
     
     Following options are experimental and only used on MacOSX:
     
@@ -395,7 +600,7 @@ class CompilationOptions:
             self.obj_only = 0
 
 
-class CompilationResult:
+class CompilationResult(object):
     """
     Results from the Cython compiler:
     
@@ -407,6 +612,7 @@ class CompilationResult:
     object_file      string or None   Result of compiling the C file
     extension_file   string or None   Result of linking the object file
     num_errors       integer          Number of compilation errors
+    compilation_source CompilationSource
     """
     
     def __init__(self):
@@ -444,8 +650,8 @@ def compile_single(source, options, full_module_name = None):
     Always compiles a single file; does not perform timestamp checking or
     recursion.
     """
-    context = Context(options.include_path)
-    return context.compile(source, options, full_module_name)
+    return run_pipeline(source, options, full_module_name)
+
 
 def compile_multiple(sources, options):
     """
@@ -458,7 +664,6 @@ def compile_multiple(sources, options):
     sources = [os.path.abspath(source) for source in sources]
     processed = set()
     results = CompilationResultSet()
-    context = Context(options.include_path)
     recursive = options.recursive
     timestamps = options.timestamps
     if timestamps is None:
@@ -466,13 +671,13 @@ def compile_multiple(sources, options):
     verbose = options.verbose or ((recursive or timestamps) and not options.quiet)
     for source in sources:
         if source not in processed:
+            # Compiling multiple sources in one context doesn't quite
+            # work properly yet.
             if not timestamps or context.c_file_out_of_date(source):
                 if verbose:
                     sys.stderr.write("Compiling %s\n" % source)
-                result = context.compile(source, options)
-                # Compiling multiple sources in one context doesn't quite
-                # work properly yet.
-                context = Context(options.include_path) # to be removed later
+
+                result = run_pipeline(source, options)
                 results.add(source, result)
             processed.add(source)
             if recursive:
@@ -519,6 +724,7 @@ def main(command_line = 0):
     else:
         options = CompilationOptions(default_options)
         sources = args
+
     if options.show_version:
         sys.stderr.write("Cython version %s\n" % Version.version)
     if options.working_path!="":
@@ -551,12 +757,14 @@ default_options = dict(
     output_file = None,
     annotate = False,
     generate_pxi = 0,
-    transforms = TransformSet(),
     working_path = "",
     recursive = 0,
     timestamps = None,
     verbose = 0,
-    quiet = 0)
+    quiet = 0,
+    pragma_overrides = {},
+    emit_linenums = False,
+)
 if sys.platform == "mac":
     from Cython.Mac.MacSystem import c_compile, c_link, CCompilerError
     default_options['use_listing_file'] = 1
