@@ -1,12 +1,17 @@
+# cython: auto_cpdef=True
 #
 #   Pyrex Parser
 #
 
+# This should be done automatically
+import cython
+cython.declare(Nodes=object, ExprNodes=object, EncodedString=object)
+
 import os
 import re
 import sys
 from types import ListType, TupleType
-from Scanning import PyrexScanner, FileSourceDescriptor
+from Cython.Compiler.Scanning import PyrexScanner, FileSourceDescriptor
 import Nodes
 import ExprNodes
 import StringEncoding
@@ -79,12 +84,9 @@ def p_simple_expr(s):
     if s.sy == 'if':
         s.next()
         test = p_or_test(s)
-        if s.sy == 'else':
-            s.next()
-            other = p_test(s)
-            return ExprNodes.CondExprNode(pos, test=test, true_val=expr, false_val=other)
-        else:
-            s.error("Expected 'else'")
+        s.expect('else')
+        other = p_test(s)
+        return ExprNodes.CondExprNode(pos, test=test, true_val=expr, false_val=other)
     else:
         return expr
         
@@ -264,6 +266,15 @@ def p_sizeof(s):
     s.expect(')')
     return node
 
+def p_yield_expression(s):
+    # s.sy == "yield"
+    pos = s.position()
+    s.next()
+    if s.sy not in ('EOF', 'NEWLINE', ')'):
+        expr = p_expr(s)
+    s.error("generators ('yield') are not currently supported")
+    return Nodes.PassStatNode(pos)
+
 #power: atom trailer* ('**' factor)*
 
 def p_power(s):
@@ -459,7 +470,7 @@ def make_slice_node(pos, start, stop = None, step = None):
     return ExprNodes.SliceNode(pos,
         start = start, stop = stop, step = step)
 
-#atom: '(' [testlist] ')' | '[' [listmaker] ']' | '{' [dictmaker] '}' | '`' testlist '`' | NAME | NUMBER | STRING+
+#atom: '(' [testlist] ')' | '[' [listmaker] ']' | '{' [dict_or_set_maker] '}' | '`' testlist '`' | NAME | NUMBER | STRING+
 
 def p_atom(s):
     pos = s.position()
@@ -468,6 +479,8 @@ def p_atom(s):
         s.next()
         if s.sy == ')':
             result = ExprNodes.TupleNode(pos, args = [])
+        elif s.sy == 'yield':
+            result = p_yield_expression(s)
         else:
             result = p_expr(s)
         s.expect(')')
@@ -475,7 +488,7 @@ def p_atom(s):
     elif sy == '[':
         return p_list_maker(s)
     elif sy == '{':
-        return p_dict_maker(s)
+        return p_dict_or_set_maker(s)
     elif sy == '`':
         return p_backquote_expr(s)
     elif sy == 'INT':
@@ -527,26 +540,22 @@ def p_atom(s):
 
 def p_name(s, name):
     pos = s.position()
-    if not s.compile_time_expr:
-        try:
-            value = s.compile_time_env.lookup_here(name)
-        except KeyError:
-            pass
+    if not s.compile_time_expr and name in s.compile_time_env:
+        value = s.compile_time_env.lookup_here(name)
+        rep = repr(value)
+        if isinstance(value, bool):
+            return ExprNodes.BoolNode(pos, value = value)
+        elif isinstance(value, int):
+            return ExprNodes.IntNode(pos, value = rep)
+        elif isinstance(value, long):
+            return ExprNodes.IntNode(pos, value = rep, longness = "L")
+        elif isinstance(value, float):
+            return ExprNodes.FloatNode(pos, value = rep)
+        elif isinstance(value, (str, unicode)):
+            return ExprNodes.StringNode(pos, value = value)
         else:
-            rep = repr(value)
-            if isinstance(value, bool):
-                return ExprNodes.BoolNode(pos, value = value)
-            elif isinstance(value, int):
-                return ExprNodes.IntNode(pos, value = rep)
-            elif isinstance(value, long):
-                return ExprNodes.IntNode(pos, value = rep, longness = "L")
-            elif isinstance(value, float):
-                return ExprNodes.FloatNode(pos, value = rep)
-            elif isinstance(value, (str, unicode)):
-                return ExprNodes.StringNode(pos, value = value)
-            else:
-                error(pos, "Invalid type for compile-time constant: %s"
-                    % value.__class__.__name__)
+            error(pos, "Invalid type for compile-time constant: %s"
+                % value.__class__.__name__)
     return ExprNodes.NameNode(pos, name = name)
 
 def p_cat_string_literal(s):
@@ -672,11 +681,11 @@ def p_string_literal(s):
     #print "p_string_literal: value =", repr(value) ###
     return kind, value
 
-# list_display  	::=  	"[" [listmaker] "]"
-# listmaker 	::= 	expression ( list_for | ( "," expression )* [","] )
-# list_iter 	::= 	list_for | list_if
-# list_for 	::= 	"for" expression_list "in" testlist [list_iter]
-# list_if 	::= 	"if" test [list_iter]
+# list_display      ::=      "[" [listmaker] "]"
+# listmaker     ::=     expression ( list_for | ( "," expression )* [","] )
+# list_iter     ::=     list_for | list_if
+# list_for     ::=     "for" expression_list "in" testlist [list_iter]
+# list_if     ::=     "if" test [list_iter]
         
 def p_list_maker(s):
     # s.sy == '['
@@ -687,16 +696,13 @@ def p_list_maker(s):
         return ExprNodes.ListNode(pos, args = [])
     expr = p_simple_expr(s)
     if s.sy == 'for':
-        loop = p_list_for(s)
+        target = ExprNodes.ListNode(pos, args = [])
+        append = ExprNodes.ComprehensionAppendNode(
+            pos, expr=expr, target=ExprNodes.CloneNode(target))
+        loop = p_list_for(s, Nodes.ExprStatNode(append.pos, expr=append))
         s.expect(']')
-        inner_loop = loop
-        while not isinstance(inner_loop.body, Nodes.PassStatNode):
-            inner_loop = inner_loop.body
-            if isinstance(inner_loop, Nodes.IfStatNode):
-                 inner_loop = inner_loop.if_clauses[0]
-        append = ExprNodes.ListComprehensionAppendNode( pos, expr = expr )
-        inner_loop.body = Nodes.ExprStatNode(pos, expr = append)
-        return ExprNodes.ListComprehensionNode(pos, loop = loop, append = append)
+        return ExprNodes.ComprehensionNode(
+            pos, loop=loop, append=append, target=target)
     else:
         exprs = [expr]
         if s.sy == ',':
@@ -705,52 +711,96 @@ def p_list_maker(s):
         s.expect(']')
         return ExprNodes.ListNode(pos, args = exprs)
         
-def p_list_iter(s):
+def p_list_iter(s, body):
     if s.sy == 'for':
-        return p_list_for(s)
+        return p_list_for(s, body)
     elif s.sy == 'if':
-        return p_list_if(s)
+        return p_list_if(s, body)
     else:
-        return Nodes.PassStatNode(s.position())
+        # insert the 'append' operation into the loop
+        return body
 
-def p_list_for(s):
+def p_list_for(s, body):
     # s.sy == 'for'
     pos = s.position()
     s.next()
     kw = p_for_bounds(s)
     kw['else_clause'] = None
-    kw['body'] = p_list_iter(s)
+    kw['body'] = p_list_iter(s, body)
     return Nodes.ForStatNode(pos, **kw)
         
-def p_list_if(s):
+def p_list_if(s, body):
     # s.sy == 'if'
     pos = s.position()
     s.next()
     test = p_test(s)
     return Nodes.IfStatNode(pos, 
-        if_clauses = [Nodes.IfClauseNode(pos, condition = test, body = p_list_iter(s))],
+        if_clauses = [Nodes.IfClauseNode(pos, condition = test,
+                                         body = p_list_iter(s, body))],
         else_clause = None )
-    
+
 #dictmaker: test ':' test (',' test ':' test)* [',']
 
-def p_dict_maker(s):
+def p_dict_or_set_maker(s):
     # s.sy == '{'
     pos = s.position()
     s.next()
-    items = []
-    while s.sy != '}':
-        items.append(p_dict_item(s))
-        if s.sy != ',':
-            break
+    if s.sy == '}':
         s.next()
-    s.expect('}')
-    return ExprNodes.DictNode(pos, key_value_pairs = items)
-    
-def p_dict_item(s):
-    key = p_simple_expr(s)
-    s.expect(':')
-    value = p_simple_expr(s)
-    return ExprNodes.DictItemNode(key.pos, key=key, value=value)
+        return ExprNodes.DictNode(pos, key_value_pairs = [])
+    item = p_simple_expr(s)
+    if s.sy == ',' or s.sy == '}':
+        # set literal
+        values = [item]
+        while s.sy == ',':
+            s.next()
+            if s.sy == '}':
+                break
+            values.append( p_simple_expr(s) )
+        s.expect('}')
+        return ExprNodes.SetNode(pos, args=values)
+    elif s.sy == 'for':
+        # set comprehension
+        target = ExprNodes.SetNode(pos, args=[])
+        append = ExprNodes.ComprehensionAppendNode(
+            item.pos, expr=item, target=ExprNodes.CloneNode(target))
+        loop = p_list_for(s, Nodes.ExprStatNode(append.pos, expr=append))
+        s.expect('}')
+        return ExprNodes.ComprehensionNode(
+            pos, loop=loop, append=append, target=target)
+    elif s.sy == ':':
+        # dict literal or comprehension
+        key = item
+        s.next()
+        value = p_simple_expr(s)
+        if s.sy == 'for':
+            # dict comprehension
+            target = ExprNodes.DictNode(pos, key_value_pairs = [])
+            append = ExprNodes.DictComprehensionAppendNode(
+                item.pos, key_expr=key, value_expr=value,
+                target=ExprNodes.CloneNode(target))
+            loop = p_list_for(s, Nodes.ExprStatNode(append.pos, expr=append))
+            s.expect('}')
+            return ExprNodes.ComprehensionNode(
+                pos, loop=loop, append=append, target=target)
+        else:
+            # dict literal
+            items = [ExprNodes.DictItemNode(key.pos, key=key, value=value)]
+            while s.sy == ',':
+                s.next()
+                if s.sy == '}':
+                    break
+                key = p_simple_expr(s)
+                s.expect(':')
+                value = p_simple_expr(s)
+                items.append(
+                    ExprNodes.DictItemNode(key.pos, key=key, value=value))
+            s.expect('}')
+            return ExprNodes.DictNode(pos, key_value_pairs=items)
+    else:
+        # raise an error
+        s.expect('}')
+    return ExprNodes.DictNode(pos, key_value_pairs = [])
 
 def p_backquote_expr(s):
     # s.sy == '`'
@@ -902,6 +952,21 @@ def p_print_statement(s):
     return Nodes.PrintStatNode(pos,
         arg_tuple = arg_tuple, append_newline = not ends_with_comma)
 
+def p_exec_statement(s):
+    # s.sy == 'exec'
+    pos = s.position()
+    s.next()
+    args = [ p_bit_expr(s) ]
+    if s.sy == 'in':
+        s.next()
+        args.append(p_simple_expr(s))
+        if s.sy == ',':
+            s.next()
+            args.append(p_simple_expr(s))
+    else:
+        error(pos, "'exec' currently requires a target mapping (globals/locals)")
+    return Nodes.ExecStatNode(pos, args = args)
+
 def p_del_statement(s):
     # s.sy == 'del'
     pos = s.position()
@@ -980,7 +1045,8 @@ def p_import_statement(s):
         else:
             if as_name and "." in dotted_name:
                 name_list = ExprNodes.ListNode(pos, args = [
-                    ExprNodes.StringNode(pos, value = EncodedString("*"))])
+                        ExprNodes.IdentifierStringNode(
+                            pos, value = EncodedString("*"))])
             else:
                 name_list = None
             stat = Nodes.SingleAssignmentNode(pos,
@@ -1005,14 +1071,20 @@ def p_from_import_statement(s, first_statement = 0):
     else:
         s.error("Expected 'import' or 'cimport'")
     is_cimport = kind == 'cimport'
+    is_parenthesized = False
     if s.sy == '*':
         imported_names = [(s.position(), "*", None, None)]
         s.next()
     else:
+        if s.sy == '(':
+            is_parenthesized = True
+            s.next()
         imported_names = [p_imported_name(s, is_cimport)]
     while s.sy == ',':
         s.next()
         imported_names.append(p_imported_name(s, is_cimport))
+    if is_parenthesized:
+        s.expect(')')
     dotted_name = EncodedString(dotted_name)
     if dotted_name == '__future__':
         if not first_statement:
@@ -1291,7 +1363,7 @@ def p_include_statement(s, ctx):
 def p_with_statement(s):
     pos = s.position()
     s.next() # 'with'
-#	if s.sy == 'IDENT' and s.systring in ('gil', 'nogil'):
+#    if s.sy == 'IDENT' and s.systring in ('gil', 'nogil'):
     if s.sy == 'IDENT' and s.systring == 'nogil':
         state = s.systring
         s.next()
@@ -1307,8 +1379,8 @@ def p_with_statement(s):
             if not allow_multi and isinstance(target, ExprNodes.TupleNode):
                 s.error("Multiple with statement target values not allowed without paranthesis")
         body = p_suite(s)
-	return Nodes.WithStatNode(pos, manager = manager, 
-	       			       target = target, body = body)
+    return Nodes.WithStatNode(pos, manager = manager, 
+                              target = target, body = body)
     
 def p_simple_statement(s, first_statement = 0):
     #print "p_simple_statement:", s.sy, s.systring ###
@@ -1316,6 +1388,8 @@ def p_simple_statement(s, first_statement = 0):
         node = p_global_statement(s)
     elif s.sy == 'print':
         node = p_print_statement(s)
+    elif s.sy == 'exec':
+        node = p_exec_statement(s)
     elif s.sy == 'del':
         node = p_del_statement(s)
     elif s.sy == 'break':
@@ -1330,6 +1404,8 @@ def p_simple_statement(s, first_statement = 0):
         node = p_import_statement(s)
     elif s.sy == 'from':
         node = p_from_import_statement(s, first_statement = first_statement)
+    elif s.sy == 'yield':
+        node = p_yield_expression(s)
     elif s.sy == 'assert':
         node = p_assert_statement(s)
     elif s.sy == 'pass':
@@ -1403,6 +1479,7 @@ def p_IF_statement(s, ctx):
 
 def p_statement(s, ctx, first_statement = 0):
     cdef_flag = ctx.cdef_flag
+    decorators = None
     if s.sy == 'ctypedef':
         if ctx.level not in ('module', 'module_pxd'):
             s.error("ctypedef statement not allowed here")
@@ -1414,63 +1491,71 @@ def p_statement(s, ctx, first_statement = 0):
     elif s.sy == 'IF':
         return p_IF_statement(s, ctx)
     elif s.sy == 'DECORATOR':
-        if ctx.level not in ('module', 'class', 'c_class', 'property'):
+        if ctx.level not in ('module', 'class', 'c_class', 'property', 'module_pxd', 'c_class_pxd'):
+            print ctx.level
             s.error('decorator not allowed here')
         s.level = ctx.level
         decorators = p_decorators(s)
-        if s.sy != 'def':
+        if s.sy not in ('def', 'cdef', 'cpdef'):
             s.error("Decorators can only be followed by functions ")
-        return p_def_statement(s, decorators)
+    elif s.sy == 'pass' and cdef_flag:
+        # empty cdef block
+        return p_pass_statement(s, with_newline = 1)
+
+    overridable = 0
+    if s.sy == 'cdef':
+        cdef_flag = 1
+        s.next()
+    elif s.sy == 'cpdef':
+        cdef_flag = 1
+        overridable = 1
+        s.next()
+    if cdef_flag:
+        if ctx.level not in ('module', 'module_pxd', 'function', 'c_class', 'c_class_pxd'):
+            s.error('cdef statement not allowed here')
+        s.level = ctx.level
+        node = p_cdef_statement(s, ctx(overridable = overridable))
+        if decorators is not None:
+            if not isinstance(node, (Nodes.CFuncDefNode, Nodes.CVarDefNode)):
+                s.error("Decorators can only be followed by functions ")
+            node.decorators = decorators
+        return node
     else:
-        overridable = 0
-        if s.sy == 'cdef':
-            cdef_flag = 1
-            s.next()
-        if s.sy == 'cpdef':
-            cdef_flag = 1
-            overridable = 1
-            s.next()
-        if cdef_flag:
-            if ctx.level not in ('module', 'module_pxd', 'function', 'c_class', 'c_class_pxd'):
-                s.error('cdef statement not allowed here')
+        if ctx.api:
+            error(s.pos, "'api' not allowed with this statement")
+        elif s.sy == 'def':
+            if ctx.level not in ('module', 'class', 'c_class', 'c_class_pxd', 'property'):
+                s.error('def statement not allowed here')
             s.level = ctx.level
-            return p_cdef_statement(s, ctx(overridable = overridable))
+            return p_def_statement(s, decorators)
+        elif s.sy == 'class':
+            if ctx.level != 'module':
+                s.error("class definition not allowed here")
+            return p_class_statement(s)
+        elif s.sy == 'include':
+            if ctx.level not in ('module', 'module_pxd'):
+                s.error("include statement not allowed here")
+            return p_include_statement(s, ctx)
+        elif ctx.level == 'c_class' and s.sy == 'IDENT' and s.systring == 'property':
+            return p_property_decl(s)
+        elif s.sy == 'pass' and ctx.level != 'property':
+            return p_pass_statement(s, with_newline = 1)
         else:
-            if ctx.api:
-                error(s.pos, "'api' not allowed with this statement")
-            elif s.sy == 'def':
-                if ctx.level not in ('module', 'class', 'c_class', 'c_class_pxd', 'property'):
-                    s.error('def statement not allowed here')
-                s.level = ctx.level
-                return p_def_statement(s)
-            elif s.sy == 'class':
-                if ctx.level != 'module':
-                    s.error("class definition not allowed here")
-                return p_class_statement(s)
-            elif s.sy == 'include':
-                if ctx.level not in ('module', 'module_pxd'):
-                    s.error("include statement not allowed here")
-                return p_include_statement(s, ctx)
-            elif ctx.level == 'c_class' and s.sy == 'IDENT' and s.systring == 'property':
-                return p_property_decl(s)
-            elif s.sy == 'pass' and ctx.level != 'property':
-                return p_pass_statement(s, with_newline = 1)
+            if ctx.level in ('c_class_pxd', 'property'):
+                s.error("Executable statement not allowed here")
+            if s.sy == 'if':
+                return p_if_statement(s)
+            elif s.sy == 'while':
+                return p_while_statement(s)
+            elif s.sy == 'for':
+                return p_for_statement(s)
+            elif s.sy == 'try':
+                return p_try_statement(s)
+            elif s.sy == 'with':
+                return p_with_statement(s)
             else:
-                if ctx.level in ('c_class_pxd', 'property'):
-                    s.error("Executable statement not allowed here")
-                if s.sy == 'if':
-                    return p_if_statement(s)
-                elif s.sy == 'while':
-                    return p_while_statement(s)
-                elif s.sy == 'for':
-                    return p_for_statement(s)
-                elif s.sy == 'try':
-                    return p_try_statement(s)
-                elif s.sy == 'with':
-                    return p_with_statement(s)
-                else:
-                    return p_simple_statement_list(
-                        s, ctx, first_statement = first_statement)
+                return p_simple_statement_list(
+                    s, ctx, first_statement = first_statement)
 
 def p_statement_list(s, ctx, first_statement = 0):
     # Parse a series of statements separated by newlines.
@@ -1590,7 +1675,7 @@ def p_calling_convention(s):
     else:
         return ""
 
-calling_convention_words = ("__stdcall", "__cdecl")
+calling_convention_words = ("__stdcall", "__cdecl", "__fastcall")
 
 def p_c_complex_base_type(s):
     # s.sy == '('
@@ -1607,17 +1692,28 @@ def p_c_simple_base_type(s, self_flag, nonempty):
     is_basic = 0
     signed = 1
     longness = 0
+    complex = 0
     module_path = []
     pos = s.position()
+    if not s.sy == 'IDENT':
+        error(pos, "Expected an identifier, found '%s'" % s.sy)
     if looking_at_base_type(s):
         #print "p_c_simple_base_type: looking_at_base_type at", s.position()
         is_basic = 1
-        signed, longness = p_sign_and_longness(s)
-        if s.sy == 'IDENT' and s.systring in basic_c_type_names:
+        if s.sy == 'IDENT' and s.systring in special_basic_c_types:
+            signed, longness = special_basic_c_types[s.systring]
             name = s.systring
             s.next()
         else:
-            name = 'int'
+            signed, longness = p_sign_and_longness(s)
+            if s.sy == 'IDENT' and s.systring in basic_c_type_names:
+                name = s.systring
+                s.next()
+            else:
+                name = 'int'
+        if s.sy == 'IDENT' and s.systring == 'complex':
+            complex = 1
+            s.next()
     elif looking_at_dotted_name(s):
         #print "p_c_simple_base_type: looking_at_type_name at", s.position()
         name = s.systring
@@ -1646,7 +1742,8 @@ def p_c_simple_base_type(s, self_flag, nonempty):
     type_node = Nodes.CSimpleBaseTypeNode(pos, 
         name = name, module_path = module_path,
         is_basic_c_type = is_basic, signed = signed,
-        longness = longness, is_self_arg = self_flag)
+        complex = complex, longness = longness, 
+        is_self_arg = self_flag)
 
 
     # Treat trailing [] on type as buffer access if it appears in a context
@@ -1734,13 +1831,19 @@ def looking_at_dotted_name(s):
         return result
     else:
         return 0
-        
-basic_c_type_names = ("void", "char", "int", "float", "double", "Py_ssize_t", "bint")
+
+basic_c_type_names = ("void", "char", "int", "float", "double", "bint")
+
+special_basic_c_types = {
+    # name : (signed, longness)
+    "Py_ssize_t" : (2, 0),
+    "size_t"     : (0, 0),
+}
 
 sign_and_longness_words = ("short", "long", "signed", "unsigned")
 
 base_type_start_words = \
-    basic_c_type_names + sign_and_longness_words
+    basic_c_type_names + sign_and_longness_words + tuple(special_basic_c_types)
 
 def p_sign_and_longness(s):
     signed = 1
@@ -1975,31 +2078,30 @@ def p_cdef_statement(s, ctx):
     elif s.sy == 'import':
         s.next()
         return p_cdef_extern_block(s, pos, ctx)
-    if p_nogil(s):
+    elif p_nogil(s):
         ctx.nogil = 1
-    if s.sy == ':':
+        if ctx.overridable:
+            error(pos, "cdef blocks cannot be declared cpdef")
+        return p_cdef_block(s, ctx)
+    elif s.sy == ':':
+        if ctx.overridable:
+            error(pos, "cdef blocks cannot be declared cpdef")
         return p_cdef_block(s, ctx)
     elif s.sy == 'class':
         if ctx.level not in ('module', 'module_pxd'):
             error(pos, "Extension type definition not allowed here")
-        #if ctx.api:
-        #    error(pos, "'api' not allowed with extension class")
+        if ctx.overridable:
+            error(pos, "Extension types cannot be declared cpdef")
         return p_c_class_definition(s, pos, ctx)
-    elif s.sy == 'IDENT' and s.systring in struct_union_or_enum:
+    elif s.sy == 'IDENT' and s.systring in ("struct", "union", "enum", "packed"):
         if ctx.level not in ('module', 'module_pxd'):
             error(pos, "C struct/union/enum definition not allowed here")
-        #if ctx.visibility == 'public':
-        #    error(pos, "Public struct/union/enum definition not implemented")
-        #if ctx.api:
-        #    error(pos, "'api' not allowed with '%s'" % s.systring)
+        if ctx.overridable:
+            error(pos, "C struct/union/enum cannot be declared cpdef")
         if s.systring == "enum":
             return p_c_enum_definition(s, pos, ctx)
         else:
             return p_c_struct_or_union_definition(s, pos, ctx)
-    elif s.sy == 'pass':
-        node = p_pass_statement(s)
-        s.expect_newline('Expected a newline')
-        return node
     else:
         return p_c_func_or_var_declaration(s, pos, ctx)
 
@@ -2007,6 +2109,8 @@ def p_cdef_block(s, ctx):
     return p_suite(s, ctx(cdef_flag = 1))
 
 def p_cdef_extern_block(s, pos, ctx):
+    if ctx.overridable:
+        error(pos, "cdef extern blocks cannot be declared cpdef")
     include_file = None
     s.expect('from')
     if s.sy == '*':
@@ -2021,10 +2125,6 @@ def p_cdef_extern_block(s, pos, ctx):
         include_file = include_file,
         body = body)
 
-struct_union_or_enum = (
-    "struct", "union", "enum"
-)
-
 def p_c_enum_definition(s, pos, ctx):
     # s.sy == ident 'enum'
     s.next()
@@ -2075,6 +2175,12 @@ def p_c_enum_item(s, items):
         name = name, cname = cname, value = value))
 
 def p_c_struct_or_union_definition(s, pos, ctx):
+    packed = False
+    if s.systring == 'packed':
+        packed = True
+        s.next()
+        if s.sy != 'IDENT' or s.systring != 'struct':
+            s.expected('struct')
     # s.sy == ident 'struct' or 'union'
     kind = s.systring
     s.next()
@@ -2100,7 +2206,7 @@ def p_c_struct_or_union_definition(s, pos, ctx):
     return Nodes.CStructOrUnionDefNode(pos, 
         name = name, cname = cname, kind = kind, attributes = attributes,
         typedef_flag = ctx.typedef_flag, visibility = ctx.visibility,
-        in_pxd = ctx.level == 'module_pxd')
+        in_pxd = ctx.level == 'module_pxd', packed = packed)
 
 def p_visibility(s, prev_visibility):
     pos = s.position()
@@ -2128,7 +2234,7 @@ def p_c_func_or_var_declaration(s, pos, ctx):
                                 assignable = 1, nonempty = 1)
     declarator.overridable = ctx.overridable
     if s.sy == ':':
-        if ctx.level not in ('module', 'c_class'):
+        if ctx.level not in ('module', 'c_class', 'module_pxd', 'c_class_pxd'):
             s.error("C function definition not allowed here")
         doc, suite = p_suite(s, Ctx(level = 'function'), with_doc = 1)
         result = Nodes.CFuncDefNode(pos,
@@ -2172,7 +2278,7 @@ def p_ctypedef_statement(s, ctx):
         ctx.api = 1
     if s.sy == 'class':
         return p_c_class_definition(s, pos, ctx)
-    elif s.sy == 'IDENT' and s.systring in ('struct', 'union', 'enum'):
+    elif s.sy == 'IDENT' and s.systring in ('packed', 'struct', 'union', 'enum'):
         if s.systring == 'enum':
             return p_c_enum_definition(s, pos, ctx)
         else:
@@ -2382,7 +2488,7 @@ def p_code(s, level=None):
             repr(s.sy), repr(s.systring)))
     return body
 
-COMPILER_DIRECTIVE_COMMENT_RE = re.compile(r"^#\s*cython:\s*([a-z]+)\s*=(.*)$")
+COMPILER_DIRECTIVE_COMMENT_RE = re.compile(r"^#\s*cython:\s*(\w+)\s*=(.*)$")
 
 def p_compiler_directive_comments(s):
     result = {}
@@ -2392,10 +2498,10 @@ def p_compiler_directive_comments(s):
             name = m.group(1)
             try:
                 value = Options.parse_option_value(str(name), str(m.group(2).strip()))
+                if value is not None: # can be False!
+                    result[name] = value
             except ValueError, e:
                 s.error(e.args[0], fatal=False)
-            if value is not None: # can be False!
-                result[name] = value
         s.next()
     return result
 
