@@ -3,12 +3,12 @@
 #           and associated know-how.
 #
 
-from Cython import Utils
 import Naming
 import PyrexTypes
+import StringEncoding
 import sys
 
-class Signature:
+class Signature(object):
     #  Method slot signature descriptor.
     #
     #  has_dummy_arg      boolean
@@ -119,7 +119,7 @@ class Signature:
         return None
 
 
-class SlotDescriptor:
+class SlotDescriptor(object):
     #  Abstract base class for type slot descriptors.
     #
     #  slot_name    string           Member name of the slot in the type object
@@ -127,13 +127,15 @@ class SlotDescriptor:
     #  flag                          Py_TPFLAGS_XXX value indicating presence of slot
     #  py3k                          Indicates presence of slot in Python 3
     #  py2                           Indicates presence of slot in Python 2
+    #  ifdef                         Full #ifdef string that slot is wrapped in. Using this causes py3k, py2 and flags to be ignored.)
 
-    def __init__(self, slot_name, dynamic = 0, flag = None, py3k = True, py2 = True):
+    def __init__(self, slot_name, dynamic = 0, flag = None, py3k = True, py2 = True, ifdef = None):
         self.slot_name = slot_name
         self.is_initialised_dynamically = dynamic
         self.flag = flag
         self.py3k = py3k
         self.py2  = py2
+        self.ifdef = ifdef
 
     def generate(self, scope, code):
         if self.is_initialised_dynamically:
@@ -143,16 +145,24 @@ class SlotDescriptor:
         flag = self.flag
         py3k = self.py3k
         py2  = self.py2
-        if not py3k:
-            code.putln("#if PY_MAJOR_VERSION < 3")
-        elif not py2:
+        if self.ifdef:
+            code.putln("#if %s" % self.ifdef)
+        else:
+            if not py3k:
+                code.putln("#if PY_MAJOR_VERSION < 3")
+            elif not py2:
+                code.putln("#if PY_MAJOR_VERSION >= 3")
+            if flag:
+                code.putln("#if (PY_MAJOR_VERSION >= 3) || (Py_TPFLAGS_DEFAULT & %s)" % flag)
+        if py3k == '<RESERVED>':
             code.putln("#if PY_MAJOR_VERSION >= 3")
-        if flag:
-            code.putln("#if (PY_MAJOR_VERSION >= 3) || (Py_TPFLAGS_DEFAULT & %s)" % flag)
+            code.putln("0, /*reserved*/")
+            code.putln("#else")
+
         code.putln("%s, /*%s*/" % (value, self.slot_name))
-        if flag:
+        if py3k == '<RESERVED>':
             code.putln("#endif")
-        if not py3k or not py2:
+        if flag or (not py3k or not py2) or self.ifdef:
             code.putln("#endif")
 
     # Some C implementations have trouble statically 
@@ -199,8 +209,8 @@ class MethodSlot(SlotDescriptor):
     #  method_name  string           The __xxx__ name of the method
     #  default      string or None   Default value of the slot
     
-    def __init__(self, signature, slot_name, method_name, default = None, flag = None, py3k=True, py2=True):
-        SlotDescriptor.__init__(self, slot_name, flag = flag, py3k = py3k, py2=py2)
+    def __init__(self, signature, slot_name, method_name, default = None, flag = None, py3k=True, py2=True, ifdef=None):
+        SlotDescriptor.__init__(self, slot_name, flag = flag, py3k = py3k, py2=py2, ifdef=ifdef)
         self.signature = signature
         self.slot_name = slot_name
         self.method_name = method_name
@@ -209,10 +219,13 @@ class MethodSlot(SlotDescriptor):
 
     def slot_code(self, scope):
         entry = scope.lookup_here(self.method_name)
-        if entry:
+        if entry and entry.func_cname:
             return entry.func_cname
-        else:
-            return "0"
+        if self.default is not None:
+            entry = scope.lookup_here(self.default)
+            if entry and entry.func_cname:
+                return entry.func_cname
+        return "0"
 
 
 class InternalMethodSlot(SlotDescriptor):
@@ -296,7 +309,7 @@ class TypeFlagsSlot(SlotDescriptor):
     #  Descriptor for the type flags slot.
     
     def slot_code(self, scope):
-        value = "Py_TPFLAGS_DEFAULT|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_BASETYPE"
+        value = "Py_TPFLAGS_DEFAULT|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_NEWBUFFER"
         if scope.needs_gc():
             value += "|Py_TPFLAGS_HAVE_GC"
         return value
@@ -311,7 +324,7 @@ class DocStringSlot(SlotDescriptor):
                 doc = scope.doc.utf8encode()
             else:
                 doc = scope.doc.byteencode()
-            return '"%s"' % Utils.escape_byte_string(doc)
+            return '__Pyx_DOCSTR("%s")' % StringEncoding.escape_byte_string(doc)
         else:
             return "0"
 
@@ -395,7 +408,7 @@ method_name_to_slot = {}
 ## extern function pointer.
 #
 #slots_initialised_from_extern = (
-#	"tp_free",
+#    "tp_free",
 #)
 
 #------------------------------------------------------------------------------------------
@@ -554,8 +567,8 @@ PyNumberMethods = (
     MethodSlot(binaryfunc, "nb_xor", "__xor__"),
     MethodSlot(binaryfunc, "nb_or", "__or__"),
     EmptySlot("nb_coerce", py3k = False),
-    MethodSlot(unaryfunc, "nb_int", "__int__"),
-    MethodSlot(unaryfunc, "nb_long", "__long__"),
+    MethodSlot(unaryfunc, "nb_int", "__int__", default="__long__"),
+    MethodSlot(unaryfunc, "nb_long", "__long__", default="__int__", py3k = "<RESERVED>"),
     MethodSlot(unaryfunc, "nb_float", "__float__"),
     MethodSlot(unaryfunc, "nb_oct", "__oct__", py3k = False),
     MethodSlot(unaryfunc, "nb_hex", "__hex__", py3k = False),
@@ -609,8 +622,8 @@ PyBufferProcs = (
     MethodSlot(getsegcountproc, "bf_getsegcount", "__getsegcount__", py3k = False),
     MethodSlot(getcharbufferproc, "bf_getcharbuffer", "__getcharbuffer__", py3k = False),
 
-    MethodSlot(getbufferproc, "bf_getbuffer", "__getbuffer__", flag = "Py_TPFLAGS_HAVE_NEWBUFFER"),
-    MethodSlot(releasebufferproc, "bf_releasebuffer", "__releasebuffer__", flag = "Py_TPFLAGS_HAVE_NEWBUFFER"),
+    MethodSlot(getbufferproc, "bf_getbuffer", "__getbuffer__", ifdef = "PY_VERSION_HEX >= 0x02060000"),
+    MethodSlot(releasebufferproc, "bf_releasebuffer", "__releasebuffer__", ifdef = "PY_VERSION_HEX >= 0x02060000")
 )
 
 #------------------------------------------------------------------------------------------
